--
--
--
--
--

local function scene(self, name)
	if self then
		return self.scenes[name and name or self.activeScene]
	end

	local scene = {
		roomWidth = 0,
		roomHeight = 0,

		queue = game:newQueue({type = "scene"}),

		entities = {},
		entityOrder = {},
		systems = {},
		systemOrder = {},
		entitiesToRemove = {},
		systemsToRemove = {},
		drawOrder = {}
	}

	function scene:getEntities(filter)
		local list = {}

		if not filter then
			filter = function(entity) return true end
		end

		for name, entity in pairs(self.entities) do
			if filter(entity) then
				table.insert(list, entity)
			end
		end

		return list
	end

	function scene:validate(entity, options, templating)
		if entity == nil then
			return false
		end

		if type(entity) == "string" then
			if love.filesystem.exists(string.format("assets/objects/%s.lua", entity)) then
				entity = require(string.format("assets.objects.%s", entity))()
			else
				error(string.format("Attempted to load object that doesn't exist: %s", entity))
			end
		end

		if not entity.validated then
			entity.x = entity.x or 0
			entity.y = entity.y or 0
			entity.z = entity.z or 0
			entity.queue = game:newQueue({type = "entity"})

			if options then
				entity = table.merge(entity, options)
			end

			for index, system in pairs(self.systemOrder) do
				if self.systems[system]:filter(entity, true) then
					entity = self.systems[system]:validate(entity, templating)
				end
			end

			entity.validated = true

		end

		return entity
	end

	function scene:add(entity, options)
		if entity == nil then
			return false
		end

		entity = self:validate(entity, options)	local name = 'entity'

		if entity.name then
			name = entity.name
		else
			if entity.class then
				local index = 0

				while self.entities[string.format('%s_%i', entity.class, index)] do
					index = index + 1
				end

				name = string.format('%s_%i', entity.class, index)
			else
				local index = 0

				while self.entities[string.format('entity_%i', index)] do
					index = index + 1
				end

				name = string.format('entity_%i', index)
			end
		end

		entity.name = name

		for index, system in pairs(self.systems) do
			if system:filter(entity) then
				system:add(entity)
			end
		end

		table.insert(self.entityOrder, entity.name)
		self.entities[entity.name] = entity

		return entity
	end

	function scene:template(entity, options)
		if entity == nil then
			return false
		end

		return self:validate(entity, options, true)
	end

	function scene:addSystem(name)
		if love.filesystem.exists(string.format("game/systems/%s.lua", name)) then
			table.insert(self.systemOrder, name)
			self.systems[name] = require(string.format("game.systems.%s", name))()
		else
			error("Attempted to add system with no file.")
		end
	end

	function scene:get(name)
		return (self.entities[name] and self.entities[name] or false)
	end

	function scene:remove(entity)
		if type(entity) == "string" then
			game:get(entity)._removed = true
		else
			entity._removed = true
		end

		table.insert(self.entitiesToRemove, entity)
	end

	function scene:update(dt, speed)
		for index, system in pairs(self.systemOrder) do
			self.systems[system]:update(dt, speed)
		end

		for index, entity in pairs(self.entityOrder) do
			local entity = self:get(entity)

			if entity and not entity._removed then
				if entity.update then
					entity:update(dt, speed)
				end

				for index, system in pairs(self.systemOrder) do
					if self.systems[system]:filter(entity) then
						self.systems[system]:process(entity, dt, speed)
					end
				end

				table.insert(self.drawOrder, entity.name)
			end
		end
	end

	function scene:draw()
		table.sort(self.drawOrder, function(a, b)
			local a = game:get(a)
			local b = game:get(b)

			if a and b then
				local ah = a.z
				local bh = b.z

				if a.sprite then
					ah = a.y + a.sprite.height
				end

				if a.animation then
					ah = a.y + a.animation.height
				end

				if b.sprite then
					bh = b.y + b.sprite.height
				end

				if b.animation then
					bh = b.y + b.animation.height
				end

				return ah < bh
			end
		end)

		for index, system in pairs(self.systemOrder) do
			self.systems[system]:draw()
		end

		for index, entity in pairs(self.drawOrder) do
			local entity = self:get(entity)

			if entity and not entity._removed then
				if entity.draw then
					entity:draw()
				end

				for index, system in pairs(self.systemOrder) do
					if self.systems[system]:filter(entity) and not entity._removed then
						self.systems[system]:render(entity)
					end
				end
			end
		end

		self.drawOrder = {}
	end

	function scene:cleanup()
		local entitiesRemoved = 0

		for index, entity in pairs(self.entitiesToRemove) do
			if type(entity) == "string" then
				entity = self:get(entity)
			end

			for index, system in pairs(self.systemOrder) do
				if self.systems[system]:filter(entity) then
					self.systems[system]:remove(entity)
				end
			end

			self.entities[entity.name] = nil

			for index, entityB in ipairs(self.entityOrder) do
				if entity.name == entityB then
					table.remove(self.entityOrder, index)
					entitiesRemoved = entitiesRemoved + 1
				end
			end
		end

		self.entitiesToRemove = {}

		if entitiesRemoved > 0 then
			printf("%i entities removed", entitiesRemoved)
		end
	end

	return scene
end

return scene
