--
--
--
--
--

--[[
    Options:
        width - width in characters
        height - height in characters
        layers - list of strings, in order drawn from bottom to top
]]

return function(options)
    local rogue = {
        rot = require 'lib.rot',
        layers = {},
        displays = {},

        width = 89,
        height = 32,

        mapWidth = 89,
        mapHeight = 32,
        map = {},
        rooms = {},

        lastLog = 0,
        log = {
            "Game begins.",
            "Press [C] to examine things."
        }
    }

    if not options then
        options = {}
    end

    for key, value in pairs(options) do
        rogue[key] = value
    end

    if #rogue.layers == 0 then
        error "DoomsdayCascader: No layers provided."
    end

    for index, layer in pairs(rogue.layers) do
        rogue.displays[layer] = rogue.rot.Display(rogue.width, rogue.height)
        rogue.displays[layer]:setDefaultBackgroundColor({0, 0, 0, 0})
        rogue.currentLayer = layer
    end

	function rogue:blankWorld(width, height, fill)
		if not fill then
			fill = 1
		end

		for y = 1, height do
			self.map[y] = {}
			for x = 1, width do
				self.map[y][x] = fill
			end
		end

		self:renderMap("map")
	end

    function rogue:getLayer(layer)
        if not layer then
            layer = self.currentLayer
        end

        return self.displays[layer]
    end

    function rogue:colerp(a, b, t)
        return self.rot.Color.interpolate(a, b, t)
    end

    function rogue:setLayer(layer)
        if self.displays[layer] then
            self.currentLayer = layer
        else
            error "DoomsdayCascader: Attempted to select a layer that doesn't exist."
        end
    end

    function rogue:print(...)
        self.displays[self.currentLayer]:write(...)
    end

    function rogue:write(...)
        return self.displays[self.currentLayer]:drawText(...)
    end

    function rogue:writeCenter(...)
        return self.displays[self.currentLayer]:writeCenter(...)
    end

    function rogue:setColor(c)
        return self.displays[self.currentLayer]:setDefaultForegroundColor(c)
    end

    function rogue:setBackground(c)
        return self.displays[self.currentLayer]:setDefaultBackgroundColor(c)
    end

    function rogue:clearCanvas(layer)
        if not layer then
            layer = self.currentLayer
        end

        self.displays[layer]:clearCanvas()
    end

    function rogue:clear(canvas, c, x, y, w, h, fg, bg)
        if not canvas then
            canvas = self.currentLayer
        end

        self.displays[canvas]:clear(c, x, y, w, h, fg, bg)
    end

    function rogue:clearAll()
        for index, display in pairs(self.layers) do
            self:clear(display)
        end
    end

    function rogue:contains(x, y)
        return x > 0 and x < self.width and y > 0 and y < self.height
    end

    function rogue:update(dt, speed)
        self.prompts = {}
        self.lastLog = self.lastLog + (dt)

        for index, display in pairs(self.layers) do
            if self.displays[display].persistent == false then
                self:clear(display)
            end
        end

        if love.keyboard.isDown("f10") and game.debugMode then
            for index, display in pairs(self.layers) do
                self.displays[display].canvas:newImageData():encode("png", display .. ".png")
            end
        end

    end

    function rogue:draw()
        self:setLayer "ui"

        local printed = 0
        for button, prompt in pairs(self.prompts) do
            printed = printed + 1
            self:print(f("[%s] %s", button, prompt), 2, game.terminal.height - printed, nil, nil, true)
        end

        local lc = game:setLayer("terminal")

        for index, display in pairs(self.layers) do
            self.displays[display]:draw()
        end

        love.graphics.setCanvas(lc)

        if game.debugMode then
            local lc = game:setLayer "debug"

            for index, room in pairs(self.rooms) do
                love.graphics.rectangle("line", (room.x - 1) * 9, (room.y - 1) *  16, room.w * 9, room.h * 16)
            end

            love.graphics.setCanvas(lc)
        end
    end

    function rogue:getTile(x, y)
        return self.map[y][x]
    end

    function rogue:getTileData(x, y)
        return tile(self.map[y][x])
    end

    function rogue:setTile(x, y, value)
        if not tile(value) then
            error("Attempted to create tile that doesn't exist: " .. value)
        else
            self.map[y][x] = value
        end
    end

    function rogue:generateWorld(width, height, callback, firstFloor)
        width = width or self.width
        height = height or self.height

        self.mapWidth = width
        self.mapHeight = height
        local world = self.rot.Map.Brogue(width, height, {
            caveChance = 0,
            roomWidth = {8, 15},
            roomHeight = {5, 15},
            crossWidth = {0,0},
            crossHeight = {0,0},
            corridorChance = 0
        })

        for y = 1, height do
            table.insert(self.map, {})
            for x = 1, width do
                table.insert(self.map[y], 0)
            end
        end

        _, self.rawRooms = world:create(function(x, y, value)
            game.terminal.map[y][x] = value + 1
        end, firstFloor)

        self.rooms = {}

        for index, room in pairs(self.rawRooms) do
            table.insert(self.rooms, {
                x = room._dims.x1,
                y = room._dims.y1,
                w = room._dims.w1,
                h = room._dims.h1
            })
        end

        for y = 1, height do
            for x = 1, width do
                callback(self.map, x, y, self.map[y][x])
            end
        end
    end

    function rogue:renderMap(layer)
        local ll = self.currentLayer
        self:setLayer(layer)

        for y = 1, self.mapHeight do
            for x = 1, self.mapWidth do
                local tileData = tile(self.map[y][x])
                if tileData then
                    if type(tileData.character) == "string" then
                        self:print(tileData.character, x, y, tileData.color, tileData.background)
                    else
                        self:print(tileData.character[math.random(1, #tileData.character)], x, y, tileData.color, tileData.background)
                    end
                else
                    error(string.format("Tile placed that has no data: %s", self.map[y][x]))
                end
            end
        end
    end

    function rogue:renderSection(layer, _x, _y, w, h)
        local ll = self.currentLayer
        self:setLayer(layer)

        for y = _y, _y + h do
            for x = _x, _x + w do
                local tileData = tile(self.map[y][x])
                if tileData then
                    self:print(tileData.character, x, y, tileData.color, tileData.background)
                else
                    error(string.format("Tile placed that has no data: %s", self.map[y][x]))
                end
            end
        end
    end

    function rogue:processWorld(callback)
        self.map = callback(self.map, self.rooms)
    end

    function rogue:populateWorld()

    end

    return rogue
end
